from __future__ import division
import math

__author__ = 'Jakob Abesser'


class Converter:
    """ Converter utitily class
    """

    note_names = dict()
    note_names['b'] = ['c', 'db', 'd', 'eb', 'e', 'f', 'gb', 'g', 'ab', 'a', 'bb', 'b']
    note_names['#'] = ['c', 'c#', 'd', 'd#', 'e', 'f', 'f#', 'g', 'g#', 'a', 'a#', 'b']

    def __init__(self):
        pass

    @staticmethod
    def pitch_to_note_name(pitch, delimiter='', accidental='b', upper_case=False):
        """ Convert MIDI pitch to note name (note spelling + octave number)
        :param pitch: (int) MIDI pitch value
        :param delimiter: (string) Delimiter between note spelling & octave number
        :param accidental: (string) Key accidental ('b' or '#')
        :return: (string) Note name
        """
        noteName = "{}{}{}".format(Converter.note_names[accidental][Converter.pitch_to_chroma(pitch)],
                                   delimiter,
                                   Converter.pitch_to_octave(pitch))
        if upper_case:
            noteName = noteName[0].upper() + noteName[1:]
        return noteName

    @staticmethod
    def pitch_to_octave(pitch):
        """ Converts MIDI pitch to octave number
        :param pitch: (int) MIDI pitch
        :return: octave: (int) Octave number
        """
        return int(math.floor(pitch/12)-1)

    @staticmethod
    def pitch_to_chroma(pitch):
        """ Convert MIDI pitch to chroma value
        :param pitch: (int) MIDI pitch
        :return: (int) Chroma value
        """
        return pitch % 12

    @staticmethod
    def pitch_to_freq(pitch):
        """ Convert MIDI pitch to fundamental frequency in Hz
        :param pitch: (int / float) MIDI pitch value
        :return: freq: (float) Fundamental frequency in Hz
        """
        return 440.*2**((pitch-69)/12)

    @staticmethod
    def freq_to_pitch(freq):
        """ Convert MIDI pitch to fundamental frequency in Hz
        :param freq (float): Frequency in Hz
        :return: pitch (float): Corresponding (fractional) MIDI pitch value
        """
        return 69 + 12*math.log(freq/440)/math.log(2)




from .tools import Tools
from .convert.converter import Converter
from .transform.transformer import Transformer
from .sisa.f0_tracking.f0_tracking_peak_tracking import F0TrackerPeakTrackingAbesserDAFX2014
from .sisa.f0_tracking.estimator import F0Tracker
from .sisa.loudness.loudness_critical_band_approximation import LoudnessCriticalBandApproximationIsolatedMonophonicTracks
from .sisa.loudness.estimator import LoudnessEstimator
from .sisa.tuning.tuning_nnls_chroma import TuningEstimatorMauch
from .sisa.tuning.estimator import TuningEstimator
from .sisa.main import ScoreInformedSoloAnalysis
from .wrapper.sonic_visualiser import SonicVisualiserWrapper




from .f0_tracking_peak_tracking import F0TrackerPeakTrackingAbesserDAFX2014

__author__ = 'Jakob Abesser'


class F0Tracker:
    """ Wrapper function for fundamental frequency tracker
    """

    def __init__(self):
        pass

    @staticmethod
    def process(samples,
                sample_rate,
                pitch,
                onset,
                duration,
                **options):

        if options['f0_tracking_method'] == 'abesser_dafx_2014':
            return F0TrackerPeakTrackingAbesserDAFX2014().process(samples,
                                                                  sample_rate,
                                                                  pitch,
                                                                  onset,
                                                                  duration,
                                                                  tuning_frequency=options['tuning_frequency'])
        else:
            raise Exception("F0 tracking method not yet implemented!")





import numpy as np

from .loudness_critical_band_approximation import LoudnessCriticalBandApproximationIsolatedMonophonicTracks
from ...tools import Tools


__author__ = 'Jakob Abesser'


class LoudnessEstimator:
    """ Wrapper class for loudness estimators
    """
    
    def __init__(self):
        pass

    @staticmethod
    def process(samples,
                sample_rate,
                pitch,
                onset,
                duration,
                **options):
        """ Call note-wise loudness estimation
        :param samples: (ndarray) Audio samples
        :param sample_rate: (float) Sampling rate
        :param pitch: (ndarray) MIDI pitch values
        :param onset: (ndarray) Note onset values in seconds
        :param duration: (ndarray) Note duration values in seconds
        :param options: (dict)
        :returns results: (dict) note-wise loudness results with keys:
            'max': peak loudness over note duration
            'meadian': median loudness over note duration
            'std': standard deviation over note duration
            'tempCentroid': temporal centroid of loudness over note duration [0,1]
            'relPeakPos': relative peak position over note duration [0,1]
        """
        if type(pitch) is not np.ndarray:
            pitch = np.array(pitch)
        if type(onset) is not np.ndarray:
            onset = np.array(onset)
        if type(pitch) is not np.ndarray:
            duration = np.array(duration)

        if options['loudness_estimation_method'] == 'critical_band_approximation':
            loudness, time_sec = LoudnessCriticalBandApproximationIsolatedMonophonicTracks.process(samples,
                                                                                                   sample_rate,
                                                                                                   **options)
        else:
            raise Exception("Loudness estimation method not implemented!")

        # aggregate loudness values over note durations
        return Tools.aggregate_framewise_function_over_notes(loudness, time_sec, onset, duration)



from scipy.signal import decimate
import numpy as np

from ...transform.transformer import Transformer

__author__ = 'Jakob Abesser'


class F0TrackerPeakTrackingAbesserDAFX2014:

    def __init__(self,
                 blocksize=2048,
                 hopsize=128,
                 zero_padding_factor=8,
                 bins_per_octave=25*12,
                 down_sampling_factor=.5,
                 lower_mag_th_db=50.,
                 pitch_margin=2,
                 verbose=True,
                 visualize=False):
        self.blocksize = float(blocksize)
        self.hopsize = float(hopsize)
        self.hopsize_correction = int(self.blocksize/self.hopsize)-1
        self.zero_padding_factor = zero_padding_factor
        self.bins_per_octave = bins_per_octave
        self.down_sample_fac = down_sampling_factor
        self.lower_mag_th_db = lower_mag_th_db
        self.pitch_margin = pitch_margin
        self.verbose = verbose
        self.visualize = visualize

        # maximum absolute deviation between neighbored frames
        self.delta_bin_max = np.round(.2*self.bins_per_octave/12.)

        # initialize
        self.samples = None
        self.sample_rate = None
        self.tuning_frequency = None
        self.num_frames = 0
        self.num_samples = 0
        self.num_notes = 0
        self.global_f0_hz = None

    def _down_sampling(self):
        """ Signal downsampling
        """
        if self.down_sample_fac != 1.:
            self.samples = decimate(self.samples, int(1/self.down_sample_fac))
            self.sample_rate *= self.down_sample_fac

    def _define_time_axis(self):
        """ Define global time axis, use frame centers for time stamps
        """
        self.time_axis_samples = np.arange(self.num_frames) * self.hopsize + .5*self.blocksize
        self.time_axis_sec = self.time_axis_samples / float(self.sample_rate)

    def process(self,
                samples,
                sample_rate,
                pitch,
                onset,
                duration,
                tuning_frequency=440.):
        """ Perform fundamental frequency tracking based on given score information (note parameters)
        :param samples: (ndarray) Monaural sample vector
        :param sample_rate: (int / float) Sample rate in Hz
        :param pitch: (list / ndarray) Note MIDI pitch values
        :param onset: (list / ndarray) Note onset times in secnods
        :param duration:  (list / ndarray) Note durations in secnods
        :param tuning_frequency: (float) Tuning frequency (e.g. previously estimated using the pymus package)
        :return: global_freq_hz: (ndarray) Frame-wise f0 values in Hz (0 where no contour is present)
        :return: time_axis_sec: (ndarray) Corresponding time stamps in seconds
        :return: contours (dict) Dict with contours parameters (size: num_contours) with following keys:
            'f0_hz': (ndarray) frame-wise f0 values in Hz
            'f0_cent_rel': (ndarray) frame-wise f0 deviations from annotated pitch in cent
            't_sec': (ndarray) Time-stamps in sec
        """

        # enforce numpy array format
        if type(pitch) is not np.ndarray:
            pitch = np.array(pitch)
        if type(onset) is not np.ndarray:
            onset = np.array(onset)
        if type(pitch) is not np.ndarray:
            duration = np.array(duration)

        self.num_notes = len(pitch)
        self.samples = samples
        self.sample_rate = sample_rate
        self.tuning_frequency = tuning_frequency

        # note offsets
        offset = np.array([onset[_] + duration[_] for _ in range(self.num_notes)])

        # downsampling
        self._down_sampling()

        self.num_samples = len(self.samples)
        self.num_frames = self.num_samples / self.hopsize

        # set global time axis
        self._define_time_axis()
        self.global_f0_hz = np.zeros(len(self.time_axis_sec))

        # iterate over all notes
        contours = []
        for n in range(self.num_notes):

            if self.verbose:
                if n % 20 == 1:
                    print('Process note {}/{} ...'.format(n, self.num_notes))

            # track f0 contour for current note
            f0_hz, f0_cent_rel, time_frame_idx = self._track_f0_contour(onset[n],
                                                                        offset[n],
                                                                        pitch[n])

            # save f0 contour
            curr_contour = dict()
            curr_contour['f0_hz'] = f0_hz
            curr_contour['f0_cent_rel'] = f0_cent_rel
            curr_contour['t_sec'] = self.time_axis_sec[time_frame_idx]
            contours.append(curr_contour)

            # save contour values to global contour
            self.global_f0_hz[time_frame_idx] = f0_hz

        return self.global_f0_hz, self.time_axis_sec, contours

    def _note_boundaries_in_global_time_axis(self, onset, offset):
        """ Find note boundaries based on onset and offset time w.r.t. global time axis
        :param onset: (float) Onset time in secnods
        :param offset: (float) Offset time in seconds
        :return: onset_frame_idx: (int) Note start frame index
        :return: offset_frame_idx: (int) Note stop frame index
        """
        start_sample = np.max([0, np.round(onset*self.sample_rate) - .5*self.blocksize])
        end_sample = np.min([self.num_samples-1, np.round(offset*self.sample_rate) + .5*self.blocksize])
        return int(round(start_sample/self.hopsize)), \
               int(round(end_sample/self.hopsize))

    def _scale_and_limit_spectrogram(self, spec):
        """ Scale and limit magnitude spectrogram
        :param spec: (2d ndarray) Magnitude spectrogram
        :return: spec_scaled: (2d ndarray) Scaled magnitude spectrogram
        """
        # scale spectrogram to [0,1]
        spec -= np.min(spec)
        spec /= np.max(spec)

        # scale to dB magnitude
        spec = 20*np.log10(spec + np.spacing(1))

        # limit to lower threshold
        spec[spec < -self.lower_mag_th_db] = -self.lower_mag_th_db

        return spec

    def _track_f0_contour(self,
                          onset,
                          offset,
                          pitch):
        """
        Score-informed note-wise f0 tracking using the reassigned spectrogram
        Reference:
            J. Abesser et al., Note-wise Fundamental Frequency Contours in Trumpet and Saxophone
            Jazz Improvisations, Proceedings of the DAFX conference, Erlangen, Germany, 2014
        Args:
            onset (float): Note onset
            offset (float): Note offset
            pitch (int): Note MIDI pitch
        Returns:
            f0_hz (ndarray): Frame-wise f0 values
            f0_cent_rel (ndarray): Frame-wise deviations of f0 from target pitch (given
                                   by note MIDI pitch and tuning frequency) in cent
            frames (ndarray): Frame indices (of global time axis, see self.time_axis_samples)
        """
        start_frame, end_frame = self._note_boundaries_in_global_time_axis(onset, offset)

        # define logarithmically spaced frequency axis
        log_f_axis_midi = np.arange(pitch-self.pitch_margin,
                                    pitch+self.pitch_margin+12./self.bins_per_octave,
                                    12./self.bins_per_octave)
        log_f_axis_hz = self.tuning_frequency*2**((log_f_axis_midi-69.)/12.)

        # compute reassigned spectrogram
        start_sample = self.time_axis_samples[start_frame]
        end_sample = self.time_axis_samples[end_frame]
        spec, _, time_frame_sec, inst_freq_hz = Transformer.reassigned_spec(self.samples[start_sample:end_sample],
                                                                            self.blocksize,
                                                                            self.hopsize,
                                                                            sample_rate=self.sample_rate,
                                                                            freq_bin_hz=log_f_axis_hz,
                                                                            n_fft=self.zero_padding_factor*self.blocksize)

        # scale & limit spectrogram
        spec = self._scale_and_limit_spectrogram(spec)

        # fundamental frequency according to pitch & tuning frequency
        f0_target_hz = self.tuning_frequency*2**((pitch-69)/12.)

        # get absolute distance to f0 for all frame-wise magnitude maxima
        bin_max = np.argmax(spec, axis=0)
        abs_dist_to_f0_bin = np.abs(f0_target_hz - log_f_axis_hz[bin_max])

        # find optimal starting frame for forwards-backwards tracking (look
        # for frame with maximum peak closest to the annotated f0 value)
        # (don't use argmin here since there could be multiple bins of interest)
        start_frame_tracking = (abs_dist_to_f0_bin == np.min(abs_dist_to_f0_bin)).nonzero()[0]

        # if we have multiple potential start positions -> take peak with highest magnitude
        if len(start_frame_tracking) > 1:
            mag = spec[bin_max[start_frame_tracking], start_frame_tracking]
            start_frame_tracking = start_frame_tracking[np.argmax(mag)]

        # perform forwards-backwards tracking based on continuity assumption
        f0_bin = self._forwards_backwards_tracking(spec, start_frame_tracking, bin_max[start_frame_tracking], self.delta_bin_max)

        # frame-wise fundamental frequency values
        f0_hz = log_f_axis_hz[f0_bin]

        # frame-wise f0 deviation from annotated pitch in cent
        f0_cent_rel = (f0_bin - self.pitch_margin*self.bins_per_octave/12.)*1200./self.bins_per_octave

        # plot tracking results over spectrogram
        if self.visualize:
            import matplotlib.pyplot as plt
            plt.matshow(spec)
            plt.hold(True)
            plt.plot(bin_max, 'r-')
            plt.show(block=False)

        return f0_hz, f0_cent_rel, np.arange(start_frame, start_frame + len(f0_hz))

    def _forwards_backwards_tracking(self, spec, idx_start, bin_start, delta_bin_max):
        """ Forwards-backwards tracking of peak contour in a given magnitude spectrogram using a start position and a
            proximity criterion
        :param spec: (2d ndarray) Magnitude spectrogram (num_bins x num_frames)
        :param idx_start: (int) Start frame index
        :param bin_start: (int) Start frequency bin
        :param delta_bin_max: (int) (onse-sided) search margin from one to the next frame
        :return: peak_bin: (ndarray) Peak bin positions (num_frames)
        """
        num_bins, num_frames = spec.shape
        peak_bin = np.zeros(num_frames)
        peak_bin[idx_start] = bin_start

        # backwards tracking
        if idx_start > 0:
            for f in range(idx_start-1, -1, -1):
                peak_bin[f] = self._find_proximate_peak(spec[:, f],
                                                        peak_bin[f+1],
                                                        delta_bin_max)

        # forwards tracking
        if idx_start < num_frames - 1:
            for f in range(idx_start+1, num_frames):
                peak_bin[f] = self._find_proximate_peak(spec[:, f],
                                                        peak_bin[f-1],
                                                        delta_bin_max)

        return peak_bin.astype(int)

    def _find_proximate_peak(self, spec, peak_bin_ref, delta_bin_max):
        """ Find peak in spectogram frame based on proximity towards peak bin from previous frame (peak tracking using
            proximity criterion)
        :param spec: (ndarray) Magnitude spectrogram frame
        :param peak_bin_ref: (int) Reference peak bin from previous frame
        :param delta_bin_max: (int) (one-sided) search margin around peak_bin_ref
        :return: peak_bin (int) Peak bin position in current frame
        """
        min_bin, max_bin = self._get_margin_bin_boundaries(peak_bin_ref,
                                                           delta_bin_max,
                                                           len(spec))
        return min_bin + np.argmax(spec[min_bin:max_bin+1])

    @staticmethod
    def _get_margin_bin_boundaries(mid_bin, bin_margin, num_bins):
        """ Find lower and upper bin boundaries based on given middle bin (mid_bin), a bin margin (bin_margin) and the
            axis boundaries (0, max_bin)
        :param mid_bin: (int) Middle bin
        :param bin_margin: (int) (one-sided) bin margin
        :param num_bins: (int) Number of bins in axis
        :return: min_bin: (int) Lower boundary bin index
        :return: max_bin: (int) Upper boundary bin index
        """
        return int(np.max((0, mid_bin-bin_margin))), int(np.min((num_bins-1, mid_bin+bin_margin)))



__author__ = 'Jakob Abesser'

import unittest
import numpy as np
import os

from ..transform.transformer import Transformer
from ..tools import Tools


class TestTransform(unittest.TestCase):
    """ Unit tests for Transformer class
    """

    def setUp(self):
        """ Load reference data exported from Matlab """
        self.data = dict()
        refData = dict()
        refData['test_stft'] = ['x', 'spec_real', 'spec_imag', 'blocksize', 'hopsize', 'NFFT']
        refData['test_reass_spec'] = ['x', 'spec', 'f', 'fs', 'if', 'blocksize', 'hopsize', 'NFFT']

        for label in refData:
            self.data[label] = dict()
            for param in refData[label]:
                self.data[label][param] = np.loadtxt(Tools.get_file_path_for_test_data(label + '_' + param + '.txt'),
                                                     delimiter=',',
                                                     dtype=float)

    def test_sftf(self):
        """ Unit test for Transformer.stft() """
        spec, timeFrameSec, freqBinHz = Transformer.stft(self.data['test_stft']['x'],
                                                         int(self.data['test_stft']['blocksize']),
                                                         int(self.data['test_stft']['hopsize']),
                                                         int(self.data['test_stft']['NFFT']))

        # check dimensions
        np.testing.assert_almost_equal(spec.real, self.data['test_stft']['spec_real'], decimal=12)
        np.testing.assert_almost_equal(spec.imag, self.data['test_stft']['spec_imag'], decimal=12)

        # check for correct values
        self.assertEqual(spec.shape[0], len(freqBinHz))
        self.assertEqual(spec.shape[1], len(timeFrameSec))

    def test_reassigned_spec(self):
        """ Unit test for Transformer.reassigned_spec() """
        spec, freqBinHz, timeFrameSec, instFreqHz = Transformer.reassigned_spec(self.data['test_reass_spec']['x'],
                                                                                self.data['test_reass_spec']['blocksize'],
                                                                                self.data['test_reass_spec']['hopsize'],
                                                                                sample_rate=self.data['test_reass_spec']['fs'],
                                                                                freq_bin_hz=self.data['test_reass_spec']['f'],
                                                                                n_fft=self.data['test_reass_spec']['NFFT'])

        # check for correct values
        np.testing.assert_equal(freqBinHz, self.data['test_reass_spec']['f'])
        np.testing.assert_almost_equal(spec, self.data['test_reass_spec']['spec'], decimal=12)
        np.testing.assert_almost_equal(instFreqHz, self.data['test_reass_spec']['if'], decimal=5)

        # check for correct values
        self.assertEqual(spec.shape[0], len(freqBinHz))


if __name__ == "__main__":
    unittest.main()



__author__ = 'Jakob Abesser'


class SonicVisualiserWrapper:
    """ Class provides wrapper functions to import / export results from Sonic Visualiser annotation layers """

    def __init__(self):
        pass

    @staticmethod
    def export_values_layer(fn_csv,
                            times,
                            values):
        """ Export values layers (timestamps & values
        :param fn_csv: (string) CSV file name
        :param times: (ndarray / list) Time stamps in seconds
        :param values: (ndarray / list) Values
        """
        assert len(times) == len(values)
        with open(fn_csv, 'w+') as f:
            for _ in range(len(times)):
                f.write("{},{}\n".format(times[_], values[_]))



# try to import libsoundfile for audio import, alternatively try loading audio via scipy or wave
try:
    import soundfile
    soundfile_found = True
except ImportError:
    from scipy.io.wavfile import read
    import wave
    soundfile_found = False

import numpy as np
import os

__author__ = 'Jakob Abesser'


class Tools:
    """ Class provides several tools for audio analysis
    """

    def __init__(self):
        pass

    @staticmethod
    def set_missing_values(options,
                           **default_frame_wise_values):
        """ Add default frame_wise_values & keys to dictionary if frame_wise_values are not set
        Args:
            options (dict): Arbitrary dictionary (e.g. containing processing options)
            default_frame_wise_values (dict): Keyword list with default frame_wise_values to be set if corresponding
                                              keys are not set in options dict
        Returns:
            options (dict): Arbitrary dictionary with added default frame_wise_values if required
        """
        for param in default_frame_wise_values.keys():
            if param not in options:
                options[param] = default_frame_wise_values[param]
        return options

    @staticmethod
    def load_wav(fn_wav,
                 mono=False):
        """ Function loads samples from WAV file. Both implementations (wave / scipy package) fail for some WAV files 
            hence we combine them.
        Args:
            fn_wav (string): WAV file name
            mono (bool): Switch if samples shall be converted to mono
        Returns:
            samples (np array): Audio samples (between [-1,1]
                                 > if stereo: (2D ndarray with DIM numSamples x numChannels),
                                 > if mono: (1D ndarray with DIM numSamples)
            sample_rate (float): Sampling frequency [Hz]
        """
        if soundfile_found:
            samples, sample_rate = soundfile.read(fn_wav)
        else:
            try:
                samples, sample_rate = Tools._load_wav_file_via_scipy(fn_wav)
            except:
                try:
                    samples, sample_rate = Tools._load_wav_file_via_wave(fn_wav)
                except:
                    raise Exception("WAV file could neither be opened using Scipy nor Wave!")

        # mono conversion
        if mono:
            if samples.ndim == 2:
                if samples.shape[1] > 1:
                    samples = np.mean(samples, axis=1)
                else:
                    samples = np.squeeze(samples)

        # scaling
        if np.max(np.abs(samples)) > 1:
            samples = samples.astype(float) / 32768.0

        return samples, sample_rate

    @staticmethod
    def _load_wav_file_via_wave(fn_wav):
        """ Load samples & sample rate from WAV file """
        fp = wave.open(fn_wav)
        num_channels = fp.getnchannels()
        num_frames = fp.getnframes()
        frame_string = fp.readframes(num_frames*num_channels)
        data = np.fromstring(frame_string, np.int16)
        samples = np.reshape(data, (-1, num_channels))

        sample_rate = float(fp.getframerate())
        return samples, sample_rate

    @staticmethod
    def _load_wav_file_via_scipy(fn_wav):
        """ Load samples & sample rate from WAV file """
        inputData = read(fn_wav)
        samples = inputData[1]
        sample_rate = inputData[0]
        return samples, sample_rate

    @staticmethod
    def aggregate_framewise_function_over_notes(frame_wise_values,
                                                time_sec,
                                                onset,
                                                duration):
        """ Aggregate a frame-wise function (e.g. loudness) over note durations to obtain note-wise features
        :param frame_wise_values: (ndarray) Frame-wise values
        :param time_sec: (ndarray) Time frame frame_wise_values in seconds
        :param onset: (ndarray) Note onset times in seconds
        :param duration: (ndarray) Note durations in seconds
        :return: result: (dict of ndarrays) Note-wise aggregation results with keys
                    'max': Maximum over note duration
                    'median': Median over note duration
                    'std': Standard deviation over note duration
                    'temp_centroid': Temporal centroid over note duration [0,1]
                    'rel_peak_pos': Position of global maximum over note duration relative to note duration [0,1]
        """
        dt = time_sec[1]-time_sec[0]
        num_notes = len(onset)

        onset_frame = (onset/dt).astype(int)
        offset_frame = ((onset+duration)/dt).astype(int)

        # initialize
        result = dict()
        result['max'] = np.zeros(num_notes)
        result['median'] = np.zeros(num_notes)
        result['std'] = np.zeros(num_notes)
        result['temp_centroid'] = np.zeros(num_notes)
        result['rel_peak_pos'] = np.zeros(num_notes)

        for n in range(num_notes):
            curr_frame_wise_values = frame_wise_values[onset_frame[n]:offset_frame[n]+1]
            num_frames_curr = len(curr_frame_wise_values)
            result['max'][n] = np.max(curr_frame_wise_values)
            result['median'][n] = np.median(curr_frame_wise_values)
            result['std'][n] = np.std(curr_frame_wise_values, ddof=1)  # same result as in Matlab
            result['temp_centroid'][n] = np.sum(np.linspace(0, 1, num_frames_curr)*curr_frame_wise_values)/np.sum(curr_frame_wise_values)
            if num_frames_curr > 1:
                result['rel_peak_pos'][n] = float(np.argmax(curr_frame_wise_values))/(num_frames_curr-1)
            else:
                result['rel_peak_pos'][n] = 0

        return result

    @staticmethod
    def quadratic_interpolation(x):
        """ Peak refinement using quadratic interpolation.
        Args:
            x (ndarray): 3-element numpy array. Central element was identified as local peak before 
                         (e.g. using numpy.argmax)
        Returns:
            peak_pos (float): Interpolated peak position (relative to central element)
            peak_val (float): Interpolated peak value
        """
        peak_pos = (x[2] - x[0])/(2*(2*x[1] - x[2] - x[0]))
        peak_val = x[1] - 0.25*(x[0]-x[2])*peak_pos
        return peak_pos, peak_val

    @staticmethod
    def moving_average_filter(x,
                              N=5):
        """ Moving average on a vector
        Args:
            x (ndarray): Input vector
            N (int): Filter length
        Returns
            y (ndarry): Filtered vector
        """
        return np.convolve(x, np.ones((N,))/N, mode='valid')

    @staticmethod
    def acf(x):
        """ Autocorrelation function
        Args:
            x (ndarray): Input function
        Returns:
            acf (ndarray): Autocorrelation function
        """
        result = np.correlate(x, x, mode='full')
        return result[result.size/2:]

    @staticmethod
    def get_file_path_for_test_data(fn):
        """ Get path for test data
        :param fn: file name
        :return: absolute path of filename in subfolder /test/data
        """
        return os.path.join(os.path.abspath(os.path.dirname(__file__)), 'test', 'data', fn)



import unittest
from ..convert.converter import Converter

__author__ = 'Jakob Abesser'


class TestConvert(unittest.TestCase):
    """ Unit tests for Convert class
    """

    def setUp(self):
        pass

    def test_pitch_to_octave(self):
        """ Unit test for pitch_to_octave() """
        self.assertEqual(Converter.pitch_to_octave(0), -1)
        self.assertEqual(Converter.pitch_to_octave(11), -1)
        self.assertEqual(Converter.pitch_to_octave(12), 0)
        self.assertEqual(Converter.pitch_to_octave(23), 0)
        self.assertEqual(Converter.pitch_to_octave(24), 1)
        self.assertEqual(Converter.pitch_to_octave(48), 3)

    def test_pitch_to_chroma(self):
        """ Unit test for pitch_to_chroma() """
        self.assertEqual(Converter.pitch_to_chroma(0), 0)
        self.assertEqual(Converter.pitch_to_chroma(3), 3)
        self.assertEqual(Converter.pitch_to_chroma(11), 11)
        self.assertEqual(Converter.pitch_to_chroma(12), 0)
        self.assertEqual(Converter.pitch_to_chroma(25), 1)

    def test_midi_pitch_to_note_name(self):
        """ Unit test for midiPitch2NoteName() """
        self.assertEqual(Converter.pitch_to_note_name(24), 'c1')
        self.assertEqual(Converter.pitch_to_note_name(24, delimiter=' '), 'c 1')
        self.assertEqual(Converter.pitch_to_note_name(24, delimiter='-'), 'c-1')
        self.assertEqual(Converter.pitch_to_note_name(25, upper_case=True), 'Db1')
        self.assertEqual(Converter.pitch_to_note_name(25, upper_case=False), 'db1')
        self.assertEqual(Converter.pitch_to_note_name(25, accidental='#'), 'c#1')
        self.assertEqual(Converter.pitch_to_note_name(25, accidental='b'), 'db1')

    def test_midi_pitch_to_freq(self):
        """ Unit test for midiPitch2Freq() """
        self.assertEqual(Converter.pitch_to_freq(69), 440.)

    def test_freq_to_midi_pitch(self):
        """ Unit test for freq2MidiPitch() """
        self.assertEqual(Converter.freq_to_pitch(440.), 69)


if __name__ == "__main__":
    unittest.main()



import os

__author__ = 'Jakob Abesser'


class TuningEstimatorMauch:
    """ Class to call the Sonic Annotator commandline tool to estimate the tuning frequency using the NNLS chroma VAMP plugin
    """

    def __init__(self):
        pass

    @staticmethod
    def process(fn_wav):
        """ Wrapper to call Sonic Annotator with NNLS-Chroma Tuning VAMP plugin to extract tuning frequency for WAV file
        :param fn_wav (string) : WAV file name
        :return: tuning_freq (float) : Tuning frequency in Hz
        """
        # generate sonic-annotator command and excecute it
        fn_csv = os.path.join(os.path.dirname(fn_wav),
                              'tuning.csv')
        # os.system('export "PATH=/Applications/SonicAnnotator/:$PATH"')
        command = "sonic-annotator -d vamp:nnls-chroma:tuning:tuning \"{}\" -w csv --csv-one-file {}".format(fn_wav,
                                                                                                             fn_csv)
        os.system(command)

        # read result (tuning frequency)
        with open(fn_csv, 'r') as f:
            content = f.readline().split(',')
        if os.path.isfile(fn_csv):
            os.remove(fn_csv)

        return float(content[3])



import numpy as np

from .f0_tracking.estimator import F0Tracker
from .tuning.estimator import TuningEstimator
from .loudness.estimator import LoudnessEstimator
from ..tools import Tools

__author__ = 'Jakob Abesser'

"""
Score-informed Solo Analysis - SISA
"""


class ScoreInformedSoloAnalysis:

    """ Class allows for score-informed analysis of melodies based on given score / transcription
        information
    """

    def __init__(self):
        pass

    def analyze(self,
                fn_wav_tuning_estimation,
                fn_wav_f0_tracking,
                fn_wav_loudness_estimation,
                pitch,
                onset,
                duration,
                do_estimate_tuning_frequency=False,
                do_estimate_f0_contours=False,
                do_estimate_loudness=False,
                **options):

        if not (do_estimate_f0_contours or do_estimate_loudness or do_estimate_tuning_frequency):
            print("No processing mode activated!")
            return None


        # spectral estimation steps
        results = dict()

        # tuning estimation
        if do_estimate_tuning_frequency:

            options = Tools.set_missing_values(options,
                                               tuning_estimation_method='mauch_nnls',
                                               fn_wav=fn_wav_tuning_estimation)

            # estimate tuning frequency
            results['tuning_frequency'] = TuningEstimator.process(**options)
        else:
            if 'tuning_frequency' not in results.keys():
                results['tuning_frequency'] = 440.

        # score-informed f0 tracking
        if do_estimate_f0_contours:

            options = Tools.set_missing_values(options,
                                               tuning_frequency=results['tuning_frequency'],
                                               f0_tracking_method='abesser_dafx_2014')

            # load file
            samples, sample_rate = Tools.load_wav(fn_wav_f0_tracking, mono=True)

            results['f0_hz'], \
            results['f0_time_sec'], \
            results['f0_contours'] = F0Tracker.process(samples,
                                                       sample_rate,
                                                       pitch,
                                                       onset,
                                                       duration,
                                                       **options)


        # score-informed loudness estimation
        if do_estimate_loudness:

            options = Tools.set_missing_values(options,
                                               loudness_estimation_method='critical_band_approximation')

            # load file
            samples, sample_rate = Tools.load_wav(fn_wav_loudness_estimation, mono=True)

            results['loudness'] = LoudnessEstimator.process(samples,
                                                            sample_rate,
                                                            pitch,
                                                            onset,
                                                            duration,
                                                            **options)

        return results





from __future__ import  division
import unittest
import numpy as np

from ..features.f0_contour_features import ContourFeatures

__author__ = 'Jakob Abesser'


class TestFeatures(unittest.TestCase):
    """ Unit tests for ContourFeatures class
    """

    def setUp(self, show_plot=False):
        """ Generate test vibrato contour
        """
        # hopsize = 512, sample rate = 44.1 kHz
        dt = 512/44100

        # contour length = 100 frames
        N = 100
        n = np.arange(N)

        # fundamental frequency = 440 Hz
        self.f_0_hz = 440

        # modulation frequency = 3 Hz
        self.f_mod_hz = 3

        # modulation lift = 4 Hz
        self.freq_hz = self.f_0_hz + 4.*np.sin(2*np.pi*n*dt*self.f_mod_hz)

        # time frames
        self.time_sec = np.arange(N)*dt

        # frequency deviation from f0 in cent
        self.freq_rel_cent = 1200*np.log2(self.freq_hz/self.f_0_hz)

        if show_plot:
            import matplotlib.pyplot as pl
            pl.figure()
            pl.plot(self.time_sec, self.freq_hz)
            pl.show(block=False)

    def test_contour_features(self):
        """ Unit test for f0_contour_features() """
        features, feature_labels = ContourFeatures().process(self.time_sec,
                                                             self.freq_hz,
                                                             self.freq_rel_cent)

        assert len(features) == len(feature_labels)

        for _ in range(len(feature_labels)):
            print("{} : {}".format(feature_labels[_], features[_]))

        # test some features
        self.assertAlmostEqual(features[feature_labels.index("mod_freq_hz")], self.f_mod_hz, places=1)
        self.assertGreater(features[feature_labels.index("mod_num_periods")], 5)
        self.assertGreater(features[feature_labels.index("mod_dom")], .5)
        self.assertLess(features[feature_labels.index("lin_f0_slope")], .01)



if __name__ == "__main__":
    unittest.main()



from __future__ import division
import numpy as np

from ...tools import Tools
from ...transform.transformer import Transformer

__author__ = 'Jakob Abesser'


class LoudnessCriticalBandApproximationIsolatedMonophonicTracks:

    def __init__(self):
        pass

    @staticmethod
    def process(samples,
                sampleRate,
                **options):
        """ Score-informed note loudness estimation from isolated instrument tracks as used in
            J. Abesser, E. Cano, K. Frieler & M. Pfleiderer: Dynamics in Jazz Improvisation -
            Score-informed Estimation and Contextual Analysis of Tone Intensities in Trumpet and Saxophone Solos,
            in Proceedings of the 9th Conference on Interdisciplinary Musicology (CIM), 2014, Berlin, Germany
        :param samples: (ndarray) Audio samples
        :param sampleRate: (float) Sampling rate
        :param options: (dict)
        :return: loudness (ndarray) Frame-wise loudness values
        :return: timeSec (ndarray) Frame values in seconds
        """
        options = Tools.set_missing_values(options,
                                           loudness_critical_band_blocksize=512,
                                           loudness_critical_band_hopsize=480)

        # STFT magnitude spectrogram
        spec, time_frame_sec, freq_bin_hz = Transformer.stft(samples,
                                                             options["loudness_critical_band_blocksize"],
                                                             options["loudness_critical_band_hopsize"],
                                                             use_frame_center_time=True)
        mag_spec = np.absolute(spec)

        # power spectrogram
        power_spec = np.square(np.true_divide(mag_spec, options["loudness_critical_band_blocksize"]))

        num_frames = mag_spec.shape[1]

        # critical band edges in Hz
        band_edges = np.array([1, 100, 200, 300, 400, 510, 630, 770, 920, 1080, 1270, 1480, 1720, 2000, 2320, 2700,
                              3150, 3700, 4400, 5300, 6400, 7700, 9500, 12000, 15500])
        num_bands = len(band_edges)-1

        # bin indices of critical bands
        band_bins = np.ceil(band_edges*options["loudness_critical_band_blocksize"]/sampleRate)-1

        # band-wise intensity values
        band_intensities = np.zeros((num_bands, num_frames))
        for b in range(num_bands):
            band_intensities[b, :] = np.sum(power_spec[band_bins[b]:band_bins[b+1]+1, :], axis=0)

        # overall loudness
        loudness = 90.302 + 10*np.log10(np.sum(band_intensities, axis=0)+np.spacing(1))

        return loudness, time_frame_sec



from .tuning_nnls_chroma import TuningEstimatorMauch

__author__ = 'Jakob Abesser'


class TuningEstimator:
    """ Wrapper class for tuning estimation algorithms
    """

    def __init__(self):
        pass

    @staticmethod
    def process(**options):
        if options['tuning_estimation_method'] == 'mauch_nnls':
            return TuningEstimatorMauch.process(options['fn_wav'])
        else:
            raise Exception("Tuning estimation method not implemented!")





import unittest
import numpy as np

from ..sisa.main import ScoreInformedSoloAnalysis
from ..tools import Tools

__author__ = 'Jakob Abesser'


class TestScoreInformedSoloAnalysis(unittest.TestCase):

    def setUp(self):
        """ Load reference data from txt files exported from Matlab via dlmwrite
        """
        self.sisa = ScoreInformedSoloAnalysis()

        self.fn_wav = Tools.get_file_path_for_test_data('test_sisa.wav')

        # load reference transcription
        self.data = dict()
        for param in ['duration', 'onset', 'pitch', 'loudness', 'loudness_median',
                      'loudness_rel_peak_pos', 'loudness_std', 'loudness_temp_centroid',
                      'tuning_freq']:
            self.data[param] = np.loadtxt(Tools.get_file_path_for_test_data(param + '.txt'),
                                          delimiter=',',
                                          dtype=float)

    def test_load_wav(self):
        """ Unit test for audio import: check samples & sample rate
        """
        samples, sample_rate = Tools.load_wav(Tools.get_file_path_for_test_data('test_transform.wav'), mono=True)
        ref_samples = np.loadtxt(Tools.get_file_path_for_test_data('test_stft_x.txt'))
        np.testing.assert_equal(samples, ref_samples)
        self.assertEqual(sample_rate, 44100)

    # def test_loudness_critical_band_approximation(self):
    #     """ Unit test for critical band approximation method  """
    #     # return
          # TODO: (state 10.12.2015) most values are perfectly equal, however 2-3 deviate
    #     results = ScoreInformedSoloAnalysis().analyze(self.fn_wav,
    #                                                   self.data["pitch"],
    #                                                   self.data["onset"],
    #                                                   self.data["duration"],
    #                                                   do_estimate_tuning_frequency=False,
    #                                                   do_estimate_f0_contours=False,
    #                                                   do_estimate_loudness=True,
    #                                                   loudness_estimation_method='critical_band_approximation')
    #     #
    #     # # test loudness results
    #     test_num_decimals = 5
    #     np.testing.assert_almost_equal(results['loudness']['max'], self.data['loudness'], decimal=test_num_decimals)
    #     np.testing.assert_almost_equal(results['loudness']['median'], self.data['loudness_median'], decimal=test_num_decimals)
    #     np.testing.assert_almost_equal(results['loudness']['std'], self.data['loudness_std'], decimal=test_num_decimals)
    #     np.testing.assert_almost_equal(results['loudness']['temp_centroid'], self.data['loudness_temp_centroid'], decimal=test_num_decimals)
    #     np.testing.assert_almost_equal(results['loudness']['rel_peal_pos'], self.data['loudness_rel_peak_pos'], decimal=test_num_decimals)


if __name__ == '__main__':
    unittest.main()





from __future__ import division
import numpy as np
from ..tools import Tools

__author__ = 'Jakob Abesser'


class ContourFeatures:
    """ Class implements several fundamental frequency contour features that can for instance be used to
        classify different pitch modulation techniques (vibrato, bend, slide, etc.)
        Reference:
    """

    def __init__(self):
        self.options = None

    def process(self, time_sec, f0_hz, f0_cent_rel, **options):
        self.options = Tools.set_missing_values(options,
                                                feat_mod_range_cent=True,
                                                feat_av_f0_dev=True,
                                                feat_abs_av_f0_dev=True,
                                                feat_iqr_cent_rel=True,
                                                feat_lin_slope_cent_rel=True,
                                                feat_tendency_hz=True,
                                                feat_modulation_hz=True,
                                                min_mod_freq_hz=0.3,
                                                max_mod_freq_hz=10,
                                                mov_av_filter_len=5)

        # feature extraction
        features = []
        feature_labels = []
        val_range = np.arange(len(f0_cent_rel))

        if options['feat_mod_range_cent']:
            features.append(np.max(f0_cent_rel)-np.min(f0_cent_rel))
            feature_labels.append('mod_range_cent')

        if options['feat_av_f0_dev']:
            features.append(np.median(f0_cent_rel))
            feature_labels.append('av_f0_dev_median')
            features.append(np.mean(f0_cent_rel))
            feature_labels.append('av_f0_dev_mean')

        if options['feat_abs_av_f0_dev']:
            features.append(np.median(np.abs(f0_cent_rel)))
            feature_labels.append('av_abs_f0_dev_median')
            features.append(np.mean(np.abs(f0_cent_rel)))
            feature_labels.append('av_abs_f0_dev_mean')

        if options['feat_iqr_cent_rel']:
            features.append(np.subtract(*np.percentile(f0_cent_rel, [75, 25])))
            feature_labels.append('iqr_f0_dev')

        if options['feat_lin_slope_cent_rel']:
            features.append(np.polyfit(val_range, f0_cent_rel, 1)[0])
            feature_labels.append('lin_f0_slope')

        if options['feat_tendency_hz']:
            curr_features, curr_labels = self.compute_tendency_features(f0_hz)
            features += curr_features
            feature_labels += curr_labels

        if options['feat_modulation_hz']:
            curr_features, curr_labels = self.compute_modulation_features(time_sec,
                                                                          f0_hz,
                                                                          max_mod_freq_hz=self.options['max_mod_freq_hz'],
                                                                          min_mod_freq_hz=self.options['min_mod_freq_hz'])
            features += curr_features
            feature_labels += curr_labels

        return features, feature_labels

    @staticmethod
    def compute_tendency_features(freq_hz):

        val_range = np.arange(len(freq_hz))
        features = []
        feature_labels = []

        if len(freq_hz) < 3:
            num_feat = 7
            return [None for _ in range(num_feat)], ['' for _ in range(num_feat)]

        # Median frequency gradient
        f_diff = np.diff(freq_hz)
        N = float(len(f_diff))
        idx_mid = int(round(N/2.))
        features.append(np.median(f_diff))
        feature_labels.append('median_freq_gradient_overal')
        features.append(np.median(f_diff[:idx_mid]))
        feature_labels.append('median_freq_gradient_first_half')
        features.append(np.median(f_diff[idx_mid:]))
        feature_labels.append('median_freq_gradient_second_half')

        # Segmentation into ascending and descending segments
        diff_sign = np.sign(f_diff)
        num_diff_sign = len(diff_sign)
        seg_sign = []
        seg_len = []
        c = 0
        curr_diff_sign = 1
        for _ in range(num_diff_sign):
            if _ == 0:
                curr_diff_sign = diff_sign[_]
                seg_sign.append(curr_diff_sign)
                c += 1
            else:
                if diff_sign[_] != curr_diff_sign or _ == num_diff_sign-1:
                    seg_len.append(c)
                    c = 0
                    if _ < num_diff_sign - 1:
                        curr_diff_sign = diff_sign[_]
                        seg_sign.append(curr_diff_sign)
                        c = 1
                else:
                    c += 1

        seg_len = np.array(seg_len)
        seg_sign = np.array(seg_sign)

        features.append(float(np.sum(seg_len[np.nonzero(seg_sign == -1.)[0]]))/num_diff_sign)
        feature_labels.append('ratio_of_descending_segments')

        features.append(float(np.sum(seg_len[np.nonzero(seg_sign == 0.)[0]]))/num_diff_sign)
        feature_labels.append('ratio_of_contant_segments')

        try:
            features.append(float(np.sum(seg_len[np.nonzero(seg_sign == 1.)[0]]))/num_diff_sign)
        except:
            features.append(None)
        feature_labels.append('ratio_of_ascending_segments')

        features.append(np.median(seg_len) / np.sum(seg_len))
        feature_labels.append('av_rel_segment_len')

        return features, feature_labels

    @staticmethod
    def compute_modulation_features(time_sec,
                                    freq_hz,
                                    apply_moving_average_filter=False,
                                    max_mod_freq_hz=10,
                                    min_mod_freq_hz=0.3):

        features = []
        feature_labels = []

        # moving average filtering
        if apply_moving_average_filter:
            freq_hz = Tools.moving_average_filter(freq_hz, N=self.options["mov_av_filter_len"])

        # remove DC part
        freq_hz -= np.mean(freq_hz)

        # modulation frequency
        mod_freq, mod_dom = ContourFeatures.estimate_modulation_frequency_using_acf(time_sec,
                                                                                    freq_hz,
                                                                                    max_mod_freq_hz=max_mod_freq_hz,
                                                                                    min_mod_freq_hz=min_mod_freq_hz)
        features.append(mod_freq)
        feature_labels.append('mod_freq_hz')

        # modulation dominance measure
        features.append(mod_dom)
        feature_labels.append('mod_dom')

        # number of modulation periods
        features.append(np.sum(np.abs(np.diff(np.diff(freq_hz) > 0))))
        feature_labels.append('mod_num_periods')

        return features, feature_labels

    @staticmethod
    def estimate_modulation_frequency_using_acf(time_sec,
                                                freq_hz,
                                                max_mod_freq_hz=10,
                                                min_mod_freq_hz=0.3):
        """ Compute modulation frequency of given frequency contour using autocorrelation function (ACF)
        Args:
            time_sec (ndarray): Time frame values in seconds
            freq_hz (ndarray): Frequency contour values in Hz
        Returns:
            mod_freq: Modulation frequency in Hz
        """
        dt = time_sec[1] - time_sec[0]
        L = len(freq_hz)

        # auto-correlation function
        acf = Tools.acf(freq_hz)

        acf -= np.min(acf)
        acf /= np.max(acf)

        # ACF lag values in seconds
        lag_sec = np.arange(L)*dt

        # focus on search val_range
        focus_bins = np.where(np.logical_and(lag_sec > 1./max_mod_freq_hz,
                                             lag_sec < 1./min_mod_freq_hz))[0]
        if len(focus_bins) == 0:
            return None, None

        focus_acf = acf[focus_bins]
        focus_lag_sec = lag_sec[focus_bins]

        # global maximum in search val_range
        max_idx = np.argmax(focus_acf)

        # refine peak position using quadratic interpolation
        if len(focus_bins) > 2 and  0 < max_idx < len(focus_acf) - 1:
            max_idx_int, peak_val = max_idx + Tools.quadratic_interpolation(focus_acf[max_idx-1:max_idx+2])

            # get modulation frequency using linear interpolation
            frac = max_idx_int % 1.
            low_idx = int(np.floor(max_idx_int))
            mod_freq = 1./(focus_lag_sec[low_idx]*frac + focus_lag_sec[low_idx+1]*(1-frac))

            # dominance measure for modulation
            mod_dom = peak_val
        else:
            max_bin = focus_bins[max_idx]
            mod_freq = 1./lag_sec[max_bin]
            mod_dom = acf[max_bin]

        return mod_freq, mod_dom




from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='pymus',
      version='0.2.3',
      long_description=readme(),
      description='Tools for audio analysis, special focus on score-informed audio analysis of instrumental / vocal solo recordings',
      url='',
      keywords='score-informed audio analysis solo recordings tuning loudness intonation music information retrieval',
      author='Jakob Abesser',
      author_email='nudelsalat@posteo.de',
      license='MIT',
      packages=['pymus'],
      include_package_data=True,
      test_suite='nose.collector',
      tests_require=['nose'],
      zip_safe=False)


from __future__ import division
import numpy as np

__author__ = 'Jakob Abesser'


class Transformer:
    """ Class implements different time-frequency transformations """

    def __init__(self):
        pass

    @staticmethod
    def stft(samples,
             blocksize,
             hopsize,
             n_fft=None,
             window='Hanning',
             sample_rate=44100,
             use_frame_center_time=False):
        """ Short-time Fourier Transform (STFT), based on re-implementation of Matlab spectrogram() function
        :param samples: (ndarray) Audio samples
        :param blocksize: (int) Blocksize in samples
        :param hopsize: (int) Hopsize in samples
        :param n_fft: (int / None) FFT size in samples (zero-padding is used if necessary)
                     -> if None, n_fft is set to blocksize (no zero-padding)
        :param window: (string) Window type, currently implemented
                        'Hanning': N-point symmetric Hanning window
                        'DiffHanning': Differentiated Hanning window
        :param sample_rate: (float / None) Sampling rate in Hz
                            If None, sample_rate is set to 44100.
        :param use_frame_center_time: (bool) Switch to use frame center as frame time, if False, frame begin is used
        :return: spec: (2d ndarray) Complex STFT spectrogram (nFrequencyBins x nTimeFrames)
        :return: time_frame_sec: (ndarray) Time frame values in seconds
        :return: freq_bin_hz: (ndarray) Frequency bin values in Hz
        """
        if n_fft is None:
            n_fft = blocksize

        # buffer signal
        buffer_mat = Transformer.buffer_signal(samples,
                                               blocksize,
                                               hopsize)
        num_frames = buffer_mat.shape[1]

        # apply windowing
        window = np.reshape(Transformer._window_function(blocksize,
                                                         window), (-1, 1))
        buffer_mat *= window

        # time frames
        if use_frame_center_time:
            time_frame_sec = (np.arange(num_frames, dtype=float)+.5)
        else:
            time_frame_sec = (np.arange(num_frames, dtype=float))
        time_frame_sec *= hopsize/sample_rate

        # frequency bins
        freq_bin_hz = np.arange(n_fft/2+1)*sample_rate/n_fft

        # STFT
        spec = np.fft.fft(buffer_mat, n=int(n_fft), axis=0)[:int(n_fft/2)+1, :]

        # compute FFTs
        return spec, time_frame_sec, freq_bin_hz

    @staticmethod
    def buffer_signal(samples,
                      blocksize,
                      hopsize):
        """ Buffer signal into overlapping or non-overlapping frames. Missing samples are filled with zeros
        :param samples: (ndarray) Sample vector
        :param blocksize: (int) Blocksize in samples
        :param hopsize: (int) Hopsize in samples
        :return: buffer: (2d ndarray) Sample buffer, dimensions: blocksize x num_frames
        """
        overlap = blocksize - hopsize
        num_samples = len(samples)
        num_frames = np.floor((num_samples-overlap)/(blocksize-overlap))

        # create sample buffer matrix with (non-)overlapping frames
        col_idx = np.arange(num_frames) * hopsize
        row_idx = np.reshape(np.arange(blocksize),
                             (-1, 1))
        index = col_idx + row_idx
        index = index.astype(int)

        return samples[index]

    @staticmethod
    def reassigned_spec(samples,
                        blocksize,
                        hopsize,
                        sample_rate=44100,
                        freq_bin_hz=None,
                        n_fft=None):
        """ Compute reassigned magnitude spectrogram by mapping STFT magnitude values to time-frequency
            bins that correspond to the local instantaneous frequency. For harmonic signals, this results
            in a spectrogram representation with a higher sparsity, i.e., sharper harmonic peaks compared
            to the STFT. This is useful for tasks such as pitch tracking & music transcription.
        :param samples: (ndarray) Audio samples
        :param blocksize: (int) Blocksize in samples
        :param hopsize: (int) Hopsize in samples
        :param sample_rate: (int) Sampling rate in Hz
        :param freq_bin_hz: (None / ndarray) Desired frequency axis (bin values in Hz) for reassigned spectrogram,
                          must be linearly-spaced or logarithmically-spaced.
                          This allows for example to define a logarithmically-spaced frequency axis which is
                          used to map the magnitudes to.
                          If None, the common linearly-spaced FFT frequency axis based on the given STFT parameters
                          is used
        :param n_fft: (int / None) FFT size in samples (zero-padding is used if necessary)
                     -> if None, n_fft is set to blocksize (no zero-padding)
        :return: spec: (2d ndarray) Reassigned magnitude spectrogram (nFrequencyBins x nTimeFrames)
        :return: freq_bin_hz: (ndarray) Frequency bin values in Hz of reassigned spectrogram (nFrequencyBins)
        :return: time_frame_sec: (ndarray) Time frame values in seconds (nTimeFrames)
        :return: inst_freq_hz: (2d ndarray) Instantaneous frequency values [Hz]  (nFrequencyBins x nTimeFrames)
        """
        # Comptute magnitude STFT spectrogram & reassigned frequency positions based on the local instantaneous
        # frequency values
        inst_freq_hz, spec_stft, time_frame_sec, freq_bin_hz_stft = Transformer._inst_freq_abe(samples,
                                                                                               blocksize,
                                                                                               hopsize,
                                                                                               sample_rate,
                                                                                               n_fft=n_fft)

        # magnitude spectrogram
        spec_stft = np.abs(spec_stft)

        # use STFT frequency axis as target frequency axis if not specified otherwise
        if freq_bin_hz is None:
            freq_bin_hz = freq_bin_hz_stft
        assert all(np.diff(freq_bin_hz) > 0)

        nBins = len(freq_bin_hz)
        num_frames = spec_stft.shape[1]
        spec = np.zeros((nBins, num_frames))
        dt = time_frame_sec[1] - time_frame_sec[0]

        # only consider instantaneous frequencies within range of target frequency axis
        min_freq_hz = freq_bin_hz[0]
        max_freq_hz = freq_bin_hz[-1]
        idx_bin_valid = np.logical_and(inst_freq_hz >= min_freq_hz, inst_freq_hz <= max_freq_hz).nonzero()
        freq_hz_reassigned = inst_freq_hz[idx_bin_valid[0], idx_bin_valid[1]]
        time_sec_reassigned = time_frame_sec[idx_bin_valid[1]]
        num_valid_entries = len(freq_hz_reassigned)

        freq_scale_spacing_type = Transformer._scale_spacing_type(freq_bin_hz)

        if freq_scale_spacing_type == 'linear':
            df = freq_bin_hz[1] - freq_bin_hz[0]
            freq_bin_frac_reassigned = (freq_hz_reassigned - min_freq_hz)/df
        elif freq_scale_spacing_type == 'logarithmic':
            bins_per_octave = np.round(1./np.log2(freq_bin_hz[2]/freq_bin_hz[1]))
            freq_bin_frac_reassigned = bins_per_octave*np.log2(freq_hz_reassigned/min_freq_hz)
        else:
            raise Exception('Target frequency axis spacing must be linear or logarithmic!')

        # no time reassignment
        time_frame_frac_reassigned = np.round(time_sec_reassigned/dt)

        # simple frequency reassignment by mapping magnitude to target frequency bin which is closest to instantaneous
        # frequency value
        # (alternative would be a fractional mapping of magnitudes to both adjacent bins, this is more time consuming
        # and does not showed improvements in past experiments)
        freq_bin_frac_reassigned = np.round(freq_bin_frac_reassigned)

        for n in range(num_valid_entries):
            spec[freq_bin_frac_reassigned[n], time_frame_frac_reassigned[n]] += spec_stft[idx_bin_valid[0][n], idx_bin_valid[1][n]]

        return spec, freq_bin_hz, time_frame_sec, inst_freq_hz

    @staticmethod
    def _scale_spacing_type(scale):
        """
        :param scale: (ndarray) Scale values (e.g. frequency scale values in Hz)
        :return: type: (string) Scale spacing type ('linear', 'logarithmic', or 'other')
        """
        assert len(scale) > 2
        tol = 1e-10
        if all(np.diff(scale, 2)) < tol:
            return 'linear'
        elif all(np.diff(np.log2(scale), 2) < tol):
            return 'logarithmic'
        return 'other'

    @staticmethod
    def _inst_freq_abe(samples,
                       blocksize,
                       hopsize,
                       sample_rate,
                       n_fft=None):
        """ Compute instantaneous frequency values based on the method proposed in
                Toshihiro Abe et al. in ICASSP'95, Eurospeech'97
        :param samples: (ndarray) Audio samples
        :param blocksize: (int) Blocksize in samples
        :param hopsize: (int) Hopsize in samples
        :param sample_rate: (int) Sampling rate in Hz
        :param n_fft: (int / None) FFT size in samples (zero-padding is used if necessary)
                     -> if None, n_fft is set to blocksize (no zero-padding)
        :return: inst_freq_hz: (2d ndarray) Instantaneous frequency values [Hz]  (nFrequencyBins x nTimeFrames)
        :return: spec: (2d ndarray) STFT spectrogram (nFrequencyBins x nTimeFrames)
                       This can be used for magnitude reassignment as done in reassigned_spec()
        :return: time_frame_sec: (ndarray) Time frame values in seconds (nTimeFrames)
        :return: freq_bin_hz: (ndarray) Frequency bin values in Hz (nFrequencyBins)
        """
        # compute 2 STFTs with Hanning and DiffHanning window
        spec, time_frame_sec, freq_bin_hz = Transformer.stft(samples,
                                                             blocksize,
                                                             hopsize,
                                                             n_fft,
                                                             'Hanning',
                                                             sample_rate=sample_rate)
        spec_diff, _, _ = Transformer.stft(samples,
                                           blocksize,
                                           hopsize,
                                           n_fft,
                                           'DiffHanning',
                                           sample_rate=sample_rate)

        # compute instantaneous frequencies, use array broadcasting, ignore N/2 + 1 frame (fs/2)
        inst_freq_hz = np.imag(spec_diff[:-1, :]/spec[:-1, :])/(2*np.pi) + np.reshape(np.arange(n_fft/2)*sample_rate/n_fft,
                                                                                      (-1, 1))
        return inst_freq_hz, spec, time_frame_sec, freq_bin_hz

    @staticmethod
    def _window_function(N, window, sample_rate=44100.):
        """ Create window functions
        :param N: (int) window width
        :param window: (string) Window type, currently implemented
                'Hanning': N-point symmetric Hanning window with first and last sample being 0 (Matlab: hann())
                'DiffHanning': Differentiated Hanning window
        :param sample_rate: (float) Sampling rate in Hz
        :return: window: (ndarray) window function
        """
        if window == 'Hanning':
            return .5*(1-np.cos(2*np.pi*np.arange(N)/(N-1)))
        elif window == 'DiffHanning':
            return -np.pi*sample_rate / (N-1) * np.sin(2*np.pi*np.arange(N)/(N-1))
        else:
            raise Exception('Non-valid value for window')
